use anyhow::Result;
use bitreader::BitReader;
use openssl::{
    bn::{BigNum, BigNumContext},
    ec::{EcGroup, EcPoint},
    sha::sha1,
};

use crate::{key::unbase24, FIELD_BYTES};

#[derive(Clone, Copy, Debug)]
struct ProductKey {
    upgrade: bool,
    serial: u32,
    hash: u32,
    signature: u64,
}

pub fn verify(
    e_curve: &EcGroup,
    base_point: &EcPoint,
    public_key: &EcPoint,
    p_key: &str,
) -> Result<bool> {
    let mut num_context = BigNumContext::new()?;

    let p_raw = unbase24(p_key);
    for byte in &p_raw {
        print!("{:02X}, ", *byte);
    }
    println!();
    let product_key = unpack(&p_raw)?;

    let p_data = product_key.serial << 1 | product_key.upgrade as u32;

    let e = BigNum::from_u32(product_key.hash)?;
    let s = BigNum::from_slice(&product_key.signature.to_be_bytes())?;
    let mut x = BigNum::new()?;
    let mut y = BigNum::new()?;

    let mut t = EcPoint::new(e_curve)?;
    let mut p = EcPoint::new(e_curve)?;
    let mut p_2 = EcPoint::new(e_curve)?;

    t.mul(e_curve, base_point, &s, &num_context)?;
    p.mul(e_curve, public_key, &e, &num_context)?;
    p_2.mul(e_curve, public_key, &e, &num_context)?;

    p.add(e_curve, &t, &p_2, &mut num_context)?;

    p.affine_coordinates(e_curve, &mut x, &mut y, &mut num_context)?;

    let mut msg_buffer: [u8; 100] = [0; 100];

    let mut x_bin = x.to_vec();
    x_bin.reverse();
    let mut y_bin = y.to_vec();
    y_bin.reverse();

    msg_buffer[0..4].copy_from_slice(&p_data.to_le_bytes());
    msg_buffer[4..4 + FIELD_BYTES].copy_from_slice(&x_bin);
    msg_buffer[4 + FIELD_BYTES..4 + FIELD_BYTES * 2].copy_from_slice(&y_bin);

    let msg_digest = sha1(&msg_buffer);

    let hash: u32 = u32::from_be_bytes(msg_digest[0..4].try_into().unwrap()) & ((1 << 28) - 1);

    Ok(hash == product_key.hash)
}

pub fn generate(
    e_curve: &EcGroup,
    base_point: &EcPoint,
    gen_order: &BigNum,
    private_key: &BigNum,
    p_serial: u32,
    p_upgrade: bool,
) -> String {
    todo!()
}

fn unpack(p_raw: &[u8]) -> Result<ProductKey> {
    const HASH_LENGTH_BITS: u8 = 28;
    const SERIAL_LENGTH_BITS: u8 = 30;

    let mut reader = BitReader::new(p_raw);
    // The signature length is unknown, but everything else is, so we can calculate it
    let signature_length_bits =
        (p_raw.len() * 8) as u8 - (HASH_LENGTH_BITS + SERIAL_LENGTH_BITS + 1);

    let p_signature = reader.read_u64(signature_length_bits)?;
    let p_hash = reader.read_u32(HASH_LENGTH_BITS)?;
    let p_serial = reader.read_u32(SERIAL_LENGTH_BITS)?;
    let p_upgrade = reader.read_bool()?;

    Ok(ProductKey {
        upgrade: p_upgrade,
        serial: p_serial,
        hash: p_hash,
        signature: p_signature,
    })
}

fn pack(p_key: ProductKey) -> Vec<u8> {
    todo!()
}

#[cfg(test)]
mod tests {
    use std::{fs::File, io::BufReader};

    use serde_json::from_reader;

    use crate::crypto::initialize_elliptic_curve;

    #[test]
    fn test() {
        // Example product key and its BINK ID
        let product_key = "D9924-R6BG2-39J83-RYKHF-W47TT";
        let bink_id = "2E";

        // Load keys.json
        let path = "keys.json";
        let file = File::open(path).unwrap();
        let reader = BufReader::new(file);
        let keys: serde_json::Value = from_reader(reader).unwrap();

        let bink = &keys["BINK"][&bink_id];

        let p = bink["p"].as_str().unwrap();
        let a = bink["a"].as_str().unwrap();
        let b = bink["b"].as_str().unwrap();
        let gx = bink["g"]["x"].as_str().unwrap();
        let gy = bink["g"]["y"].as_str().unwrap();
        let kx = bink["pub"]["x"].as_str().unwrap();
        let ky = bink["pub"]["y"].as_str().unwrap();

        let (e_curve, gen_point, pub_point) = initialize_elliptic_curve(p, a, b, gx, gy, kx, ky);

        assert!(super::verify(&e_curve, &gen_point, &pub_point, product_key).unwrap());
    }
}
